import numpy as np
import pandas as pd
import torch
from torch.utils.data import DataLoader, Subset

from . import Dataset

device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")


class MIT_BIH_PytorchDataset(torch.utils.data.Dataset):
    """MIT BIH ECG pytorch dataset."""

    def __init__(self, csv_file):
        """Args:
        - csv_file (string): Path to the csv file."""
        self.data = pd.read_csv(csv_file, header = None)
        self.data = self.data.sample(frac=1, random_state=1).reset_index(drop=True) #SY: shuffle dataset

    def __len__(self):
        return len(self.data)

    def __getitem__(self, idx):
        x = torch.Tensor(self.data.iloc[idx, :-1].values)
        x = torch.nn.functional.normalize(x, p=1.0, dim=-1)
        y = torch.Tensor([self.data.loc[idx, 187]]).squeeze(-1).type(torch.LongTensor)
        return x, y


class MIT_BIH_Dataset(Dataset):
    def __init__(
        self,
        max_teacher_samples: int = None,
        max_student_train_queries: int = 10000,
        num_workers: int = 1,
    ):
        super().__init__(max_teacher_samples, max_student_train_queries, num_workers)
        self.input_shape = (187,)

    def prepare_data(self):
        self.train_data = MIT_BIH_PytorchDataset(
            csv_file="data/MIT-BIH/mitbih_train.csv"
        )
        self.test_data = MIT_BIH_PytorchDataset(
            csv_file="data/MIT-BIH/mitbih_test.csv"
        )
        
    def set_class_weight(self): #SY
        target = self.train_data.data[187]
        class_sample_count = np.unique(target, return_counts=True)[1]
        self.class_weight = torch.from_numpy(1. / class_sample_count).type(torch.FloatTensor)

    def set_student_train_labels(
        self, new_labels: np.ndarray, not_none_label_indexes: np.ndarray
    ):
        self.test_data.data.loc[not_none_label_indexes, 187] = new_labels
        self.student_train_samples_idxs = not_none_label_indexes

    def get_student_train_dataloader(self, batch_size: int):
        student_train_data = Subset(self.test_data, self.student_train_samples_idxs)
        student_train_loader = DataLoader(
            student_train_data, batch_size=batch_size, num_workers=self.num_workers
        )
        return student_train_loader

    def get_student_val_dataloader(self, batch_size: int) -> DataLoader:
        subset_data = Subset(self.test_data, self.student_val_idxs)
        return DataLoader(subset_data, batch_size, num_workers=self.num_workers)
